package db;
/*
 * @(#)ConnectionPool
 *
 * Copyright (c) 1998 Karl Moss. All Rights Reserved.
 *
 * You may study, use, modify, and distribute this software for any
 * purpose provided that this copyright notice appears in all copies.
 *
 * This software is provided WITHOUT WARRANTY either expressed or
 * implied.
 *
 * @author  Karl Moss
 * @version 1.0
 * @date    11Mar98
 *
 * Modified By xiaoming liu
 * @date 02Feb01
 */

import java.sql.*;
import java.util.*;
/**
 * <p>This class serves as a JDBC connection repository. Since
 * creating database connections is one of the most time
 * intensive aspects of JDBC, we'll create a pool of connections.
 * Each connection can be used and then replaced back into the
 * pool.
 *
 * <p>A properties file 'ConnectionPool.cfg' will be used to
 * specify the types of JDBC connections to create, as well as
 * the minimum and maximum size of the pool. The format of
 * the configuration file is:
 *
 *   #(comment)
 *   JDBCDriver=<JDBC driver name>
 *   JDBCConnectionURL=<JDBC Connection URL>
 *   ConnectionPoolSize=<minimum size of the pool>
 *   ConnectionPoolMax=<maximum size of the pool, or -1 for none>
 *   ConnectionUseCount=<maximum usage count for one connection>
 *   ConnectionTimeout=<maximum idle lifetime (in seconds) of
 *        a connection>
 *   <other property for JDBC connection>=<value>
 *
 * <p>Any number of additional properties may be given (such
 * as username and password) as required by the JDBC driver.
 *
 */

public class ConnectionPool
  implements TimerListener
{
  // JDBC Driver name
  String m_JDBCDriver;
  public String userID = null;

  // JDBC Connection URL
  String m_JDBCConnectionURL;

  // Minimum size of the pool
  int m_ConnectionPoolSize;

  // Maximum size of the pool
  int m_ConnectionPoolMax;

  // Maximum number of uses for a single connection, or -1 for
  // none
  int m_ConnectionUseCount;

  // Maximum connection idle time (in seconds)
  int m_ConnectionTimeout;

  // Additional JDBC properties
  java.util.Properties m_JDBCProperties;

  // The Connection pool. This is a vector of ConnectionObject
  // objects
  java.util.Vector m_pool;

  // The maximum number of simultaneous connections as reported
  // by the JDBC driver
  int m_MaxConnections = -1;

  // Our Timer object
  Timer m_timer;

    // In-use Timeout Flag, if this field is set to true (typically in search engine), the long time in-user connection will be closed
 
    boolean timeout=false;
    //database type
    public static final int MYSQL=1;
    public static final int ORACLE=0;

  /**
    * <p>Initializes the ConnectionPool object using
    * 'ConnectionPool.cfg' as the configuration file
    *
    * @return true if the ConnectionPool was initialized
    * properly
    */
  public boolean initialize() throws Exception
    {
      return initialize("./conf/ConnectionPool.cfg");
    }

  /**
    * <p>Initializes the ConnectionPool object with the specified
    * configuration file
    *
    * @param config Configuration file name
    * @return true if the ConnectionPool was initialized
    * properly
    */
  public boolean initialize(String config) throws Exception
    {
      // Load the configuration parameters. Any leftover parameters
      // from the configuration file will be considered JDBC
      // connection attributes to be used to establish the
      // JDBC connections
      boolean rc = loadConfig(config);

      if (rc) {
        // Properties were loaded; attempt to create our pool
        // of connections
        createPool();

        // Start our timer so we can timeout connections. The
        // clock cycle will be 20 seconds
        m_timer = new Timer(this, 20);
        m_timer.start();
      }
      return rc;
    }

  /**
    * <p>Destroys the pool and it's contents. Closes any open
    * JDBC connections and frees all resources
    */
  public void destroy()
    {
      try {

        // Stop our timer thread
        if (m_timer != null) {
          m_timer.stop();
          m_timer = null;
        }

        // Clear our pool
        if (m_pool != null) {

          // Loop throught the pool and close each connection
          for (int i = 0; i < m_pool.size(); i++) {
            close((ConnectionObject) m_pool.elementAt(i));
          }
        }
        m_pool = null;
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }

  /**
    * <p>Gets an available JDBC Connection. Connections will be
    * created if necessary, up to the maximum number of connections
    * as specified in the configuration file.
    *
    * @return JDBC Connection, or null if the maximum
    * number of connections has been exceeded
    */
  public synchronized java.sql.Connection getConnection()
    {
      // If there is no pool it must have been destroyed
      if (m_pool == null) {
        return null;
      }

      java.sql.Connection con = null;
      ConnectionObject connectionObject = null;
      int poolSize = m_pool.size();

      // Get the next available connection
      for (int i = 0; i < poolSize; i++) {

        // Get the ConnectionObject from the pool
        ConnectionObject co = (ConnectionObject)
          m_pool.elementAt(i);

        // If this is a valid connection and it is not in use,
        // grab it
        if (co.isAvailable()) {
          connectionObject = co;
          break;
        }
      }

      // No more available connections. If we aren't at the
      // maximum number of connections, create a new entry
      // in the pool
      if (connectionObject == null) {
        if ((m_ConnectionPoolMax < 0) ||
            ((m_ConnectionPoolMax > 0) &&
             (poolSize < m_ConnectionPoolMax))) {
          
          // Add a new connection.
          int i = addConnection();
          
          // If a new connection was created, use it
          if (i >= 0) {
            connectionObject = (ConnectionObject)
              m_pool.elementAt(i);
          }
        }
        else {
          trace("Maximum number of connections exceeded");
        }
      }

      // If we have a connection, set the last time accessed,
      // the use count, and the in use flag
      if (connectionObject != null) {
        connectionObject.inUse = true;
        connectionObject.useCount++;
        touch(connectionObject);
        con = connectionObject.con;
      }
      
      return con;
    }

  /**
    * <p>Places the connection back into the connection pool,
    * or closes the connection if the maximum use count has
    * been reached
    *
    * @param Connection object to close
    */
  public synchronized void close(java.sql.Connection con)
    {
      // Find the connection in the pool
      int index = find(con);
      
      if (index != -1) {
        ConnectionObject co = (ConnectionObject)
          m_pool.elementAt(index);
        
        // If the use count exceeds the max, remove it from
        // the pool.
        if ((m_ConnectionUseCount > 0) &&
            (co.useCount >= m_ConnectionUseCount)) {
          removeFromPool(index);
        }
        else {
          // Clear the use count and reset the time last used
          touch(co);
          co.inUse = false;
        }
      }
    }
  
    public void setInuseTimeout(){
	timeout=true;
    }
  /**
    * <p>Prints the contents of the connection pool to the
    * standard output device
    */
  public String printPool()
    {
	String result="ConnectionPool<p>";
      System.out.println("--ConnectionPool--");
      if (m_pool != null) {
        for (int i = 0; i < m_pool.size(); i++) {
          ConnectionObject co = (ConnectionObject)
            m_pool.elementAt(i);
          System.out.println("" + i + "=" + co);
	result=result+" "+i+"="+co+"<p>";
        }
      }
	return result;
    }
  
  /**
    * <p>Removes the ConnectionObject from the pool at the
    * given index
    *
    * @param index Index into the pool vector
    */
  private synchronized void removeFromPool(int index)
    {
      // Make sure the pool and index are valid
      if (m_pool != null) {

        if (index < m_pool.size()) {

          // Get the ConnectionObject and close the connection
          ConnectionObject co = (ConnectionObject)
            m_pool.elementAt(index);
          close(co);

          // Remove the element from the pool
          m_pool.removeElementAt(index);
        }
      }
    }
  
  /**
    * <p>Closes the connection in the given ConnectionObject
    *
    * @param connectObject ConnectionObject
    */
  private void close(ConnectionObject connectionObject)
    {
      if (connectionObject != null) {
        if (connectionObject.con != null) {
          try {
            
            // Close the connection
            connectionObject.con.close();
          }
          catch (Exception ex) {
            // Ignore any exceptions during close
          }

          // Clear the connection object reference
          connectionObject.con = null;
        }
      }
    }
  
  /**
    * <p>Loads the given configuration file.  All global
    * properties (such as JDBCDriver) will be
    * read and removed from the properties list. Any leftover
    * properties will be returned. Returns null if the
    * properties could not be loaded
    *
    * @param name Configuration file name
    * @return true if the configuration file was loaded
    */
  private boolean loadConfig(String name)
    throws Exception
    {
      boolean rc = false;

      // Get our class loader
      ClassLoader cl = getClass().getClassLoader();

      // Attempt to open an input stream to the configuration file.
      // The configuration file is considered to be a system
      // resource.
      java.io.InputStream in;
      
      if (cl != null) {
        in = cl.getResourceAsStream(name);
      }
      else {
        in = ClassLoader.getSystemResourceAsStream(name);
      }

      // If the input stream is null, then the configuration file
      // was not found
      if (in == null) {
        throw new Exception("ConnectionPool configuration file '" +
                           name + "' not found");
      }
      else {
        try {
          m_JDBCProperties = new java.util.Properties();
          
          // Load the configuration file into the properties table
          m_JDBCProperties.load(in);

          // Got the properties. Pull out the properties that we
          // are interested in
          m_JDBCDriver = consume(m_JDBCProperties, "JDBCDriver");

          m_JDBCConnectionURL = consume(m_JDBCProperties,
                                        "JDBCConnectionURL");
          m_ConnectionPoolSize = consumeInt(m_JDBCProperties,
                                            "ConnectionPoolSize");
          m_ConnectionPoolMax = consumeInt(m_JDBCProperties,
                                           "ConnectionPoolMax");
          m_ConnectionTimeout = consumeInt(m_JDBCProperties,
                                           "ConnectionTimeout");
          m_ConnectionUseCount = consumeInt(m_JDBCProperties,
                                            "ConnectionUseCount");
          rc = true;
        }
        finally {
          // Always close the input stream
          if (in != null) {
            try {
              in.close();
            }
            catch (Exception ex) {
            }
          }
        }
      }
      return rc;
    }

  /**
    * <p>Consumes the given property and returns the value.
    *
    * @param properties Properties table
    * @param key Key of the property to retrieve and remove from
    * the properties table
    * @return Value of the property, or null if not found
    */
  private String consume(java.util.Properties p, String key)
    {
      String s = null;

      if ((p != null) &&
          (key != null)) {

        // Get the value of the key
        s = p.getProperty(key);
        
        // If found, remove it from the properties table
        if (s != null) {
          p.remove(key);
        }
      }
      return s;
    }

  /**
    * <p>Consumes the given property and returns the integer
    * value.
    *
    * @param properties Properties table
    * @param key Key of the property to retrieve and remove from
    * the properties table
    * @return Value of the property, or -1 if not found
    */
  private int consumeInt(java.util.Properties p, String key)
    {
      int n = -1;

      // Get the String value
      String value = consume(p, key);

      // Got a value; convert to an integer
      if (value != null) {
        try {
          n = Integer.parseInt(value);
        }
        catch (Exception ex) {
        }
      }
      return n;
    }

  /**
    * <p>Creates the initial connection pool. A timer thread
    * is also created so that connection timeouts can be
    * handled.
    *
    * @return true if the pool was created
    */
  private void createPool() throws Exception
    {
      // Sanity check our properties
      if (m_JDBCDriver == null) {
        throw new Exception("JDBCDriver property not found");
      }
      if (m_JDBCConnectionURL == null) {
        throw new Exception("JDBCConnectionURL property not found");
      }
      if (m_ConnectionPoolSize < 0) {
        throw new Exception("ConnectionPoolSize property not found");
      }
      if (m_ConnectionPoolSize == 0) {
        throw new Exception("ConnectionPoolSize invalid");
      }
      if (m_ConnectionPoolMax < m_ConnectionPoolSize) {
        trace("WARNING - ConnectionPoolMax is invalid and will " +
              "be ignored");
        m_ConnectionPoolMax = -1;
      }
      if (m_ConnectionTimeout < 0) {
        // Set the default to 30 minutes
        m_ConnectionTimeout = 30;
      }

      // Dump the parameters we are going to use for the pool.
      // We don't know what type of servlet environment we will
      // be running in - this may go to the console or it
      // may be redirected to a log file
      trace("JDBCDriver = " + m_JDBCDriver);
      trace("JDBCConnectionURL = " + m_JDBCConnectionURL);
      trace("ConnectionPoolSize = " + m_ConnectionPoolSize);
      trace("ConnectionPoolMax = " + m_ConnectionPoolMax);
      trace("ConnectionUseCount = " + m_ConnectionUseCount);
      trace("ConnectionTimeout = " + m_ConnectionTimeout +
            " seconds");

      // Also dump any additional JDBC properties
      java.util.Enumeration enum = m_JDBCProperties.keys();
      while (enum.hasMoreElements()) {
        String key = (String) enum.nextElement();
        String value = m_JDBCProperties.getProperty(key);
	if (!key.equals("Password")){
	    trace("(JDBC Property) " + key + " = " + value);
	}
      }

      // Attempt to create a new instance of the specified
      // JDBC driver. Well behaved drivers will register
      // themselves with the JDBC DriverManager when they
      // are instantiated
      trace("Registering " + m_JDBCDriver);
      java.sql.Driver d = (java.sql.Driver)
        Class.forName(m_JDBCDriver).newInstance();
      
      // Create the vector for the pool
      m_pool = new java.util.Vector();

      // Bring the pool to the minimum size
      fillPool(m_ConnectionPoolSize);
    }

  /**
    * <p>Adds a new connection to the pool
    *
    * @return Index of the new pool entry, or -1 if an
    * error has occurred
    */
  private int addConnection()
    {
      int index = -1;

      try {
        // Calculate the new size of the pool
        int size = m_pool.size() + 1;

        // Create a new entry
        fillPool(size);

        // Set the index pointer to the new connection if one
        // was created
        if (size == m_pool.size()) {
          index = size - 1;
        }
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
      return index;
    }
  
  /**
    * <p>Brings the pool to the given size
    */
  private synchronized void fillPool(int size) throws Exception
    {
      boolean useProperties = true;

      String password = null;
      
      // If the only properties present are the user id and
      // password, get the connection using them instead of
      // the properties object
      if (m_JDBCProperties != null) {
          
        // Make sure there are only 2 properties, and they are
        // the user id and password
        if (m_JDBCProperties.size() == 2) {
          userID =
            getPropertyIgnoreCase(m_JDBCProperties, "user");
          password =
            getPropertyIgnoreCase(m_JDBCProperties, "password");

          // If all we've got is a user id and password then
          // don't use the properties
          if ((userID != null) && (password != null)) {
            useProperties = false;
          }
        }
        
      }
        
      // Loop while we need to create more connections
      while (m_pool.size() < size) {
        
        ConnectionObject co = new ConnectionObject();

        // Create the connection
        if (useProperties) {
          co.con = DriverManager.getConnection(m_JDBCConnectionURL,
                                              m_JDBCProperties);
        }
        else {
          co.con = DriverManager.getConnection(m_JDBCConnectionURL,
                                               userID, password);
        }
        
        // Do some sanity checking on the first connection in
        // the pool
        if (m_pool.size() == 0) {

          // Get the maximum number of simultaneous connections
          // as reported by the JDBC driver
          java.sql.DatabaseMetaData md = co.con.getMetaData();
          m_MaxConnections = md.getMaxConnections();
        }

        // Give a warning if the size of the pool will exceed
        // the maximum number of connections allowed by the
        // JDBC driver
        if ((m_MaxConnections > 0) &&
            (size > m_MaxConnections)) {
          trace("WARNING: Size of pool will exceed safe maximum of " +
                m_MaxConnections);
        }

        // Clear the in use flag
        co.inUse = false;
        
        // Set the last access time
        touch(co);
        
        m_pool.addElement(co);
      }
    }

  /**
   * Gets a the named propery, ignoring case. Returns null if
   * not found
   * @param p The property set
   * @param name The property name
   * @return The value of the propery, or null if not found
   */
  private String getPropertyIgnoreCase(java.util.Properties p,
                                       String name)
    {
      if ((p == null) || (name == null)) return null;
      
      String value = null;
      
      // Get an enumeration of the property names
      java.util.Enumeration enum = p.propertyNames();
      
      // Loop through the enum, looking for the given property name
      while (enum.hasMoreElements()) {
        String pName = (String) enum.nextElement();
        if (pName.equalsIgnoreCase(name)) {
          value = p.getProperty(pName);
          break;
        }
      }
      
      return value;
    }
  
  /**
    * <p>Find the given connection in the pool
    *
    * @return Index into the pool, or -1 if not found
    */
  private int find(java.sql.Connection con)
    {
      int index = -1;

      // Find the matching Connection in the pool
      if ((con != null) &&
          (m_pool != null)) {
        for (int i = 0; i < m_pool.size(); i++) {
          ConnectionObject co = (ConnectionObject)
            m_pool.elementAt(i);
          if (co.con == con) {
            index = i;
            break;
          }
        }
      }
      return index;
    }
  
  /**
    * <p>Called by the timer each time a clock cycle expires.
    * This gives us the opportunity to timeout connections
    */
  public synchronized void TimerEvent(Object object)
    {
      // No pool means no work
      if (m_pool == null) {
        return;
      }

      // Get the current time in milliseconds
      long now = System.currentTimeMillis();
      
      // Check for any expired connections and remove them
      for (int i = m_pool.size() - 1; i >= 0; i--) {
        ConnectionObject co = (ConnectionObject)
          m_pool.elementAt(i);

        // If the connection is not in use and it has not been
        // used recently, remove it
        if (!co.inUse) {
          if ((m_ConnectionTimeout > 0) &&
              (co.lastAccess +
               (m_ConnectionTimeout * 1000) < now)) {
            removeFromPool(i);
          }
        }
      }

       // Check for any connections that remain open for a long time (20 minutes) and remove them
      if (timeout){
	  for (int i = m_pool.size() - 1; i >= 0; i--) {
	      ConnectionObject co = (ConnectionObject)
		  m_pool.elementAt(i);

	      // If the connection is in use and it has been
	      // used for a long time, remove it
	      if (co.inUse) {
		  if ((co.lastAccess +
		       (20 * 60 *  1000) < now)) {
		      removeFromPool(i);
		  }
	      }
	  }
      }

      // Remove any connections that are no longer open
      for (int i = m_pool.size() - 1; i >= 0; i--) {
        ConnectionObject co = (ConnectionObject)
          m_pool.elementAt(i);
        try {
          // If the connection is closed, remove it from the pool
          if (co.con.isClosed()) {
            trace("Connection closed unexpectedly");
            removeFromPool(i);
          }
        }
        catch (Exception ex) {
        }
      }

      // Now ensure that the pool is still at it's minimum size
      try {
        if (m_pool != null) {
          if (m_pool.size() < m_ConnectionPoolSize) {
            fillPool(m_ConnectionPoolSize);
          }
        }
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }

  /**
    * <p>Sets the last access time for the given ConnectionObject
    */
  private void touch(ConnectionObject co)
    {
      if (co != null) {
        co.lastAccess = System.currentTimeMillis();
      }
    }

  /**
    * <p>Trace the given string
    */
  private void trace(String s)
    {
	
      System.out.println("ConnectionPool: " + s);
    }

  public int getDBType(){
      if (m_JDBCDriver.indexOf("oracle")!=-1)
           return ORACLE;
     else if (m_JDBCDriver.indexOf("mysql")!=-1)
            return MYSQL;
     return 0;
     }
}

// This package-private class is used to represent a single
// connection object
class ConnectionObject
{
  // The JDBC Connection
  public java.sql.Connection con;

  // true if this connection is currently in use
  public boolean inUse;

  // The last time (in milliseconds) that this connection was used
  public long lastAccess;

  // The number of times this connection has been used
  public int useCount;
// The time that this connnection is created
public long created = 
       System.currentTimeMillis();   ;
  /**
    * <p>Determine if the connection is available
    *
    * @return true if the connection can be used
    */
  public boolean isAvailable()
    {
      boolean available = false;

      try {

        // To be available, the connection cannot be in use
        // and must be open
        if (con != null) {
          if (!inUse &&
              !con.isClosed()) {
            available = true;
          }
        }
      }
      catch (Exception ex) {
      }
      
      return available;
    }
  
  /**
    * <p>Convert the object contents to a String
    */
  public String toString()
    {
      return "Connection=" + con +",created="+new java.util.Date(created)+ ",inUse=" + inUse +
        ",lastAccess=" + new java.util.Date(lastAccess) + ",useCount=" + useCount;
    }


}
